/* $Id: icq_v2.h 324 2002-08-04 21:01:42Z kuhlmann $
 *
 *  Defines for version 2 of the ICQ protocol.
 */

#define SRV_ACK            	0x000A
#define SRV_NEW_UIN        	0x0046
#define SRV_LOGIN_REPLY    	0x005A
#define SRV_BAD_PASSWORD   	0x0064
#define SRV_USER_ONLINE    	0x006E
#define SRV_USER_OFFLINE   	0x0078
#define SRV_QUERY          	0x0082
#define SRV_USER_FOUND     	0x008C
#define SRV_END_OF_SEARCH  	0x00A0
#define SRV_NEW_USER       	0x00B4
#define SRV_UPDATE_EXT     	0x00C8
#define SRV_RECV_MESSAGE   	0x00DC
#define SRV_X2             	0x00E6
#define SRV_INFO_REPLY     	0x0118
#define SRV_EXT_INFO_REPLY 	0x0122
#define SRV_TCP_REQUEST		0x015E
#define SRV_STATUS_UPDATE  	0x01A4
#define SRV_SYSTEM_MESSAGE 	0x01C2
#define SRV_UPDATE_SUCCESS 	0x01E0
#define SRV_X1             	0x021C
