/* $Id: icq_v5.h 2105 2006-04-17 21:48:26Z kuhlmann $
 *
 *  Defines for version 5 of the ICQ protocol.
 */

#define SRV_GO_AWAY             0x0028
#define SRV_BAD_PASS            0x0064
#define SRV_NOT_CONNECTED       0x00F0
#define SRV_TRY_AGAIN           0x00FA
#define SRV_SYS_DELIVERED_MESS  0x0104
#define SRV_UPDATE_FAIL         0x01EA
#define SRV_AUTH_UPDATE         0x01F4
#define SRV_MULTI_PACKET        0x0212
#define SRV_RAND_USER           0x024E
#define SRV_META_USER           0x03DE

#define META_SUCCESS            10
#define META_FAIL               ??
#define META_READONLY           30
